import unittest
import doctest
import nimbstor.stor
import nimbstor.tar

test_all = unittest.TestSuite()
test_all.addTest(doctest.DocTestSuite(nimbstor.stor))
test_all.addTest(doctest.DocTestSuite(nimbstor.tar, optionflags = doctest.ELLIPSIS))



#!/usr/pkg/bin/python2.6

BLOCK_SIZE = 1048576 * 4

import os, sys, hashlib, time, bz2, base64
try: import cPickle as pickle
except: import pickle
from itertools import ifilter
from nimbstor.optimized import rolling_search, stor_buffer

compress_function_bz2_method = lambda b: bz2.compress(b, 9)
compress_function_methods = {}
def compress_function(compression, meta):
  if compression: compression = tuple(compression)
  else: compression = (None, 0)
  if meta: return compress_function_bz2_method
  elif compress_function_methods.has_key(compression):
    return compress_function_methods[compression]
  elif compression == ('lzma', 10):
    import lzma
    compress_function_methods[compression] = lambda b: lzma.compress(b, {'level': 9, 'extreme': True})
  elif compression[0] == 'lzma':
    import lzma
    compress_function_methods[compression] = lambda b: lzma.compress(b, {'level': compression[1]})
  elif compression[0] == 'gzip':
    import zlib
    compress_function_methods[compression] = lambda b: zlib.compress(b, compression[1])
  elif compression[0] == 'bzip2':
    compress_function_methods[compression] = lambda b: bz2.compress(b, compression[1])
  elif compression[0] == None:
    compress_function_methods[compression] = lambda b: b
  else: raise SystemError("Unsupported compression type %s" % repr(compression))
  return compress_function_methods[compression]

decompress_function_bz2_method = lambda b: bz2.decompress(b)
decompress_function_methods = {}
def decompress_function(compression, meta):
  if compression: compression = tuple(compression)
  else: compression = (None, 0)
  if meta: return decompress_function_bz2_method
  elif decompress_function_methods.has_key(compression[0]):
    return decompress_function_methods[compression[0]]
  elif compression[0] == 'lzma':
    import lzma
    decompress_function_methods[compression[0]] = lambda b: lzma.decompress(b)
  elif compression[0] == 'gzip':
    import zlib
    decompress_function_methods[compression[0]] = lambda b: zlib.decompress(b)
  elif compression[0] == 'bzip2':
    decompress_function_methods[compression[0]] = lambda b: bz2.decompress(b)
  elif compression[0] == None:
    decompress_function_methods[compression[0]] = lambda b: b
  else: raise SystemError("Unsupported compression type %s" % repr(compression))
  return decompress_function_methods[compression[0]]

aes_keysetup_aes = None
def aes_keysetup(password):
  global aes_keysetup_aes
  if aes_keysetup_aes == None:
    import aes
    aes_keysetup_aes = aes.Keysetup
  return aes_keysetup_aes(password)

def message_encode(buffer, password, meta, compression, bufnum, checksum1, checksum2):
  """Internal function to encode message.

  Args:
    buffer (str): Data buffer.
    password (str): The password as SHA256 hash value.
    meta (boolean): Meta blocks are handled somewhat different.
    compression (tuple): Compression type to use.
    bufnum (int): Current buffer number
    checksum1 (int): Rolling checksum.
    checksum2 (str): Checksum as returned by stor_stream.checksum(buffer)
  """
  buffer = apply(compress_function(compression, meta), [buffer])
  bufrest = len(buffer) % 16
  if bufrest > 0: buffer = buffer + "\0" * (16 - bufrest)
  if password: buffer = aes_keysetup(password).encrypt(buffer)
  return (buffer, bufnum, checksum1, checksum2)

def message_decode(buffer, password, cksumpwd, meta, compression, bufnum, checksum1, checksum2, bufflen):
  """Internal function to decode message.

  Args:
    buffer (str): Data buffer.
    password (str): The password as SHA256 hash value.
    cksumpwd (str): The password for checksum (MD5 hash value).
    meta (boolean): Meta blocks are handled somewhat different.
    compression (tuple): Compression type to use.
    bufnum (int): Current buffer number
    checksum1 (int): Rolling checksum.
    checksum2 (str): Checksum as returned by stor_stream.checksum(buffer)
    bufflen (str): Correct data length to return or 0.
  """
  try:
    if password: buffer = aes_keysetup(password).decrypt(buffer)
    buffer = apply(decompress_function(compression, meta), [buffer])
    if bufflen != 0 and len(buffer) > bufflen: buffer = buffer[:bufflen]
    ctx = hashlib.sha256()
    if cksumpwd != None: ctx.update(cksumpwd)
    ctx.update(buffer)
    newcksum = base64.urlsafe_b64encode(ctx.digest())[:-1].upper()
    if newcksum != checksum2: raise SystemError()
  except Exception, err:
    raise SystemError("Wrong password or invalid data (%s)" % repr(err))
  return buffer

class stor_stream:
  """Base class for storage streams."""

  def __init__(self, password = None, verbose = False):
    """Initialize base stream.

    Args:
      password (str): Optional password to enable encryption.

    >>> stor_stream('asdf')._pass == stor_stream('qwer')._pass
    False
    """
    self._closed = False
    self._pos = 0
    self._known_blocks = {}
    self._known_parts = {}
    self._known_metadata = []
    self._backend = None
    if password != None:
      self._aes = hashlib.sha256(password).digest()
      self._pass = hashlib.md5(password).digest()
    else:
      self._aes = None
      self._pass = None
    self._archive = None
    self._verbose = verbose

  def verbose(self):
    return self._verbose

  def open(self, backend):
    """Initialize and open the stream.

    This function is used in backends to connect the stor_stream with an
    backend.
    """
    self._backend = backend

  def read_buffer(self, number, checksum1, checksum2):
    """Forward read_buffer to the backend."""
    return self._backend.read_buffer(number, checksum1, checksum2)

  def write_buffer(self, buffer, number, checksum1, checksum2):
    """Forward write_buffer to the backend."""
    self._backend.write_buffer(buffer, number, checksum1, checksum2)

  def checksum(self, data):
    """Calculate checksum for given data.
    
    Args:
      data (str): Data buffer.

    Checksum is calculated on data and the password, given in constructor.
    """
    ctx = hashlib.sha256()
    if self._pass != None: ctx.update(self._pass)
    ctx.update(data)
    return base64.urlsafe_b64encode(ctx.digest())[:-1].upper()

  def tell(self):
    """Return the current position in stream."""
    return self._pos

  def archive(self):
    """Tell the archive name.

    The archive name (checksum) is on input streams immediatly available, but
    on output streams is only available after close.
    """
    return self._archive

  def append_block_info(self, number, checksum1, checksum2, backend):
    """Register block info.

    Args:
      number (int): Block number.
      checksum1 (int): Rolling checksum.
      checksum2 (str): Block checksum as returned by self.checksum(data)
      backend (object): Backend reference, which can read this block.

    Backends use this function to register known blocks.
    """
    checksum1 = int(checksum1, 16)
    if number == 0 and checksum2 not in self._known_metadata:
      self._known_metadata.append(checksum2)
    elif checksum1 == 0: self._known_parts[checksum2] = number
    else: self._known_blocks.setdefault(checksum1, {})[checksum2] = number

  def close(self, dont_commit = False):
    """Close the stream.

    The backends close will be also called.
    """
    if self._closed: return
    if self._backend != None:
      self._backend.close(dont_commit)

  def blocks(self):
    return [(0, "%08X" % 0, self._archive_checksum, 0)] + self._metadata['blocks'][:]

  def description(self): return self._metadata['description']
  def keywords(self): return self._metadata['keywords']
  def parent(self): return self._metadata['parent']
  def compression(self): return self._metadata['compression']
  def size(self): return self._metadata['size']
  def usage(self): return self._metadata['usage']

class stor_output(stor_stream):
  """Storage output stream."""

  def __init__(self, password, compression, parent, description, keywords, parlimit = 1, onlyifnew = False, verbose = False):
    """Initialization of output stream.

    Args:
      password (str): Password for encryption (None to disable encryption).
      compression ((str, int)): Compression type as tupple of (method, level).
      parent (str): Name of the parent archive or None.
      description (str): Description of archive.
      keywords (list): List of keyword strings.
      onlyifnew (boolean): Set to True to commit empty (fully deduplicated) archives.
    """
    stor_stream.__init__(self, password, verbose)
    self._metadata = {
	'description': description,
	'keywords': keywords,
	'metainfo': [],
	'timestamp': time.time(),
	'parent': parent,
	'compression': compression,
	'blocks': [],
	'size': 0,
    }
    if parlimit > 1:
      self._procslimit = parlimit
      import multiprocessing
      self._procspool = multiprocessing.Pool(parlimit)
      self._procslist = []
    else: self._procslimit = None
    self._buffer = stor_buffer()
    self._bufnum = 1
    self._bufsize = BLOCK_SIZE * 2
    self._checksum2 = None
    self._size = 0
    self._something_written = False
    self._onlyifnew = onlyifnew
    self._compression = compression

  def write(self, buffer):
    """Write function.

    Use this function to write data to the stream.
    """
    buflen = len(buffer)
    self._pos += buflen
    self._buffer.push(buffer)
    self.flush_buffer(self._bufsize)

  def set_metainfo(self, info):
    """Set the meta information, user defined."""
    self._metadata['metainfo'] = info

  def set_parent(self, parent):
    """Set the parent archive name."""
    self._metadata['parent'] = parent

  def _block_already_exists(self, data, pos, checksum1):
    """Checks if the given block already exists.

    Args:
      data (str): Data buffer.
      pos (int): Block start position.
      checksum1 (int): Rolling checksum of the data.

    Call only if the block with given rolling checksum really exist (it means
    that it was registred with append_block_info).
    """
    checksum2 = self.checksum(data)
    #print "@ block found: %08X (%s)" % (checksum1, checksum2)
    if self._known_blocks[checksum1].has_key(checksum2):
      self._checksum2 = checksum2
      self._checksumblock = self._known_blocks[checksum1][checksum2]
      if self._verbose:
        print "@ deduplication:", checksum2
      return True
    return False

  def close(self, dont_commit = False):
    """Close the output stream.

    This function write also the metadata, without metadata the archive does
    not exist.
    """
    if self._closed: return
    try:
      try:
        if self._backend != None and dont_commit != True:
          self.flush_buffer(BLOCK_SIZE)
          self.commit_buffer(self._buffer.pop(), 0)
          if not self._onlyifnew or self._something_written:
            self._finish_write(0)
            self._metadata['usage'] = self._size
            self._metadata['size'] = self._pos
            buffer = pickle.dumps(self._metadata)
            self._bufnum = 0
            self.commit_buffer(buffer, 0, meta = True)
            self._finish_write(0)
      finally: stor_stream.close(self, dont_commit)
    finally:
      if self._procslimit != None:
        self._procspool.terminate()
        self._procspool.join()

  def flush_buffer(self, tosize):
    """Flush data from buffer.
    
    This function do the deduplication with rolling checksum.
    """
    while len(self._buffer) >= tosize:
      buffer = self._buffer.value()
      pos, cks = rolling_search(self._known_blocks, self._block_already_exists, buffer, BLOCK_SIZE)
      if pos != None:
        if pos > 0: self.commit_buffer(buffer[:pos], 0)
        data = buffer[pos:tosize]
        self._metadata['blocks'].append((self._checksumblock, "%08X" % cks, self._checksum2, BLOCK_SIZE))
        self._bufnum += 1
        self._found_buffer = None
        self._checksum2 = None
        self._buffer.truncate(pos + BLOCK_SIZE)
      else: self.commit_buffer(self._buffer.pop(BLOCK_SIZE), cks)

  def _finish_write(self, limit):
    if self._procslimit != None:
      while len(self._procslist) > limit:
        proc = self._procslist.pop(0)
        buffer, bufnum, checksum1, checksum2 = proc.get()
        self._size += len(buffer)
        self._backend.write_buffer(buffer, bufnum, checksum1, checksum2)

  def commit_buffer(self, buffer, checksum1int, meta = False):
    """Write data to the beckend.

    Args:
      buffer (str): Block buffer.
      checksum1int (int): Rolling checksum of the buffer.
      meta (boolean): Set to True if the buffer contains metadata.
    """
    checksum1 = "%08X" % checksum1int
    checksum2 = self.checksum(buffer)
    buflen = len(buffer)
    if checksum1int == 0 and self._known_parts.has_key(checksum2):
      self._metadata['blocks'].append((self._known_parts[checksum2], checksum1, checksum2, buflen))
    else:
      self._something_written = True
      msgencode_args = [
          buffer,
          self._aes,
          meta,
          self._compression,
          self._bufnum,
          checksum1,
          checksum2,
      ]
      if self._procslimit != None:
        self._procslist.append(self._procspool.apply_async(message_encode, msgencode_args))
        self._finish_write(self._procslimit)
      else:
        buffer, bufnum, checksum1, checksum2 = apply(message_encode, msgencode_args)
        self._size += len(buffer)
        self._backend.write_buffer(buffer, bufnum, checksum1, checksum2)
      self._metadata['blocks'].append((self._bufnum, checksum1, checksum2, buflen))
    if not meta: self._bufnum += 1
    else: self._archive = checksum2

class stor_input(stor_stream):
  """Storage input stream."""

  def __init__(self, password, archive, parlimit = 1, verbose = False):
    """Initialization of input stream.

    Args:
      password (str): Password for decrypting data (or None).
      archive (str): Name of the archive (checksum).
    """
    stor_stream.__init__(self, password, verbose)
    if parlimit > 1:
      import multiprocessing
      self._procspool = multiprocessing.Pool()
      self._procslist = []
      self._procslimit = parlimit
    else: self._procslimit = None
    self._archive_checksum = archive
    self._buffer = stor_buffer(splitsize = 10240)
    self._archive = archive
    self._prof = 0

  def open(self, backend):
    """Initialize backend for reading.

    To read data, we need to read the metadata first. This function should be
    called by backend, if the backend is ready.
    """
    stor_stream.open(self, backend)
    if self._archive_checksum not in self._known_metadata:
      raise SystemError("Archive %s not found." % repr(self._archive_checksum))
    ck1 = "%08X" % 0
    ck2 = self._archive_checksum
    try:
      data = self.read_buffer(0, ck1, ck2)
      data = message_decode(data, self._aes, self._pass, True, None, 0, ck1, ck2, 0)
      self._metadata = pickle.loads(data)
    except: raise SystemError('Failed to read metdata')
    self._compression = self._metadata['compression']
    self._blocks = self._metadata['blocks'][:]

  def close(self, dont_commit = False):
    if self._closed: return
    try: stor_stream.close(self, dont_commit)
    finally:
      if self._procslimit != None:
        self._procspool.terminate()
        self._procspool.join()

  def read(self, size):
    """Read data of given size.

    Use this function to read the data.
    """
    data = None
    if self._procslimit != None:
      while len(self._buffer) < size and (len(self._blocks) > 0 or len(self._procslist) > 0):
        while len(self._procslist) < self._procslimit and len(self._blocks) > 0:
          num, ck1, ck2, blen = self._blocks.pop(0)
          data = self.read_buffer(num, ck1, ck2)
          self._procslist.append(
            self._procspool.apply_async(message_decode, [
              data, self._aes, self._pass, False,
              self._compression, num, ck1, ck2, blen]))
        data = self._procslist.pop(0).get()
        self._buffer.push(data)
    else:
      while len(self._buffer) < size and len(self._blocks) > 0:
        num, ck1, ck2, blen = self._blocks.pop(0)
        data = self.read_buffer(num, ck1, ck2)
        data = apply(message_decode, [
              data, self._aes, self._pass, False,
              self._compression, num, ck1, ck2, blen])
        self._buffer.push(data)
    x = time.time()
    data = self._buffer.pop(size)
    self._prof += time.time() - x
    return data

  def metainfo(self):
    """Returns user defined metadata."""
    return self._metadata['metainfo']

class stor_util(stor_stream):
  """Storage stream utility.

  This storage type is used to manage the archives, f.e. to search, copy,
  remove, ...
  """

  def __init__(self, password, verbose = False):
    stor_stream.__init__(self, password, verbose)
    self._errors = []
    self._metadata = None
    self._metadata_data = {}

  def open(self, backend):
    stor_stream.open(self, backend)
    checksum1 = "%08X" % 0
    for metadata in self._known_metadata:
      data = None
      try:
        data = self.read_buffer(0, checksum1, metadata)
        data = message_decode(data, self._aes, self._pass, True, None, 0, checksum1, metadata, 0)
        data = pickle.loads(data)
      except Exception, err:
        self._errors.append((metadata, err))
        continue
      if data != None:
        self._metadata_data[metadata] = data

  def search(self, keywords):
    """Search for an archive.

    Args:
      keywords (list): List of strings, for which should be searched.

    The search function search for an archive, gives each archive weight (how
    good it matches with given keywords), sort it in ascend order and return
    the list of archvies, the dependecies between archives and list of errors
    on search.
    """
    result = []; parents = {}
    for metadata, data in self._metadata_data.items():
      if len(keywords) > 0:
        weight = 0
        for kwd in keywords:
          for fnd in ifilter(lambda x: x == kwd, data['keywords']): weight += 10
          for fnd in ifilter(lambda x: kwd in x and x != kwd, data['keywords']): weight += 3
          if kwd in time.strftime("%F %R", time.localtime(data['timestamp'])): weight += 2
          if kwd in str(data['size']): weight += 1
          if kwd.lower() in data['description'].lower(): weight += 5
          if kwd in repr(data['metainfo']): weight += 1
        result.append({
          'weight':       weight,
          'id':           metadata,
          'description':  data['description'],
          'keywords':     data['keywords'],
          'size':         data['size'],
          'usage':        data['usage'],
          'timestamp':    data['timestamp'],
          'parent':       data['parent'],
        })
      else:
        result.append({
          'weight':       0,
          'id':           metadata,
          'description':  data['description'],
          'keywords':     data['keywords'],
          'size':         data['size'],
          'usage':        data['usage'],
          'timestamp':    data['timestamp'],
          'parent':       data['parent'],
        })
      parents.setdefault(data['parent'], []).append(metadata)
    if len(keywords) > 0: result.sort(key = lambda x: x['weight'])
    else: result.sort(key = lambda x: x['description'])
    return (result, parents, self._errors)

  def keywordlist(self):
    """Return list of used keywords."""
    keywords = {}
    for metadata, data in self._metadata_data.items():
      for kwd in data['keywords']:
	keywords[kwd] = keywords.get(kwd, 0) + 1
    return (keywords, self._errors)

  def block_exists(self, number, checksum1, checksum2):
    if type(checksum1) != type(0):
      checksum1 = int(checksum1, 16)
    if number == 0:
      if checksum2 in self._known_metadata:
        return True
    elif checksum1 == 0:
      if self._known_parts.has_key(checksum2):
        if number == self._known_parts[checksum2]:
          return True
    else:
      if self._known_blocks.has_key(checksum1):
        if self._known_blocks[checksum1].has_key(checksum2):
          if number == self._known_blocks[checksum1][checksum2]:
            return True
    return False

  def description(self, archive):
    return self._metadata_data[archive]['description']

  def keywords(self, archive):
    return self._metadata_data[archive]['keywords']

  def parent(self, archive):
    return self._metadata_data[archive]['parent']

  def compression(self, archive):
    return self._metadata_data[archive]['compression']

  def size(self, archive):
    return self._metadata_data[archive]['size']

  def usage(self, archive):
    return self._metadata_data[archive]['usage']

  def blocks(self, archive):
    self._metadata = self._metadata_data[archive]
    self._archive_checksum = archive
    try: ret = stor_stream.blocks(self)
    finally:
      self._metadata = None
      self._archive_checksum = None
    return ret

  def verify_blocks(self, archive):
    self._metadata = self._metadata_data[archive]
    self._archive_checksum = archive
    if self._verbose:
      print "VERIFY ARCHIVE:", archive
    try:
      for num, cks1, cks2, size in stor_stream.blocks(self):
        self.read_buffer(num, cks1, cks2)
        if self._verbose:
          print "OK %08d.%s.%s" % (num, cks1, cks2)
    finally:
      self._metadata = None
      self._archive_checksum = None

__all__ = ('stor_output', 'stor_input', 'stor_util')
# vim:sts=2:sw=2:



#!/usr/pkg/bin/python2.6 -O

""" Prepare data for tests.


  >>> os.system('rm -rf f1 f2 f3 out1 out2 out3 .nimbstor_parent || true')
  0
  >>> os.mkdir('out1')
  >>> os.mkdir('out2')
  >>> os.mkdir('out3')
  >>> seq1 = "".join([chr(random.randint(40, 60)) for x in xrange(1024 + 7777)] * 127)
  >>> seq2 = "".join([chr(random.randint(40, 60)) for x in xrange(1024 + 7777)] * 127)
  >>> seq3 = "".join([chr(random.randint(40, 60)) for x in xrange(1024 + 7777)] * 127)
  >>> open('f1', 'w+').write(seq1)
  >>> open('f2', 'w+').write(seq2)
  >>> open('f3', 'w+').write(seq3)
  >>> opt = command_tar_arguments(['test', '-v', '-p', 'qwer', '-m', '10', '-d', 'out1', 'out2', 'out3', '-D', 'test1', '-K', 'kwd1', '-c', 'f1', 'f2', 'f3'])
  >>> command_create_archive(opt, time.time())
  Use following backends: out1, out2, out3
  f1
  f2
  f3
  Archive ...

"""

import os, sys, time, random, getpass, re
from tarfile import TarFile, PAX_FORMAT, open as taropen, filemode, CHRTYPE, BLKTYPE, SYMTYPE, LNKTYPE, DIRTYPE
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from nimbstor import stor_input, stor_output, stor_util

class NimbTarFile(TarFile):
  def __init__(self, verbose, *options, **keywords):
    TarFile.__init__(self, *options, **keywords)
    self._recorded_files_list = []
    self._verbose = verbose

  def addfile(self, tarinfo, fileobj = None):
    self._recorded_files_list.append(tarinfo.get_info('UTF-8', 'replace'))
    tarinfo.name = re.sub(r'^[./]*', '', os.path.normpath(tarinfo.name))
    if self._verbose:
      print tarinfo.name
    return TarFile.addfile(self, tarinfo, fileobj)

  def recorded_files(self):
    return self._recorded_files_list

def create_archive(output, files, verbose = False):
  tar = NimbTarFile(
      verbose = verbose,
      fileobj = output,
      mode = 'w',
      encoding = PAX_FORMAT)
  try:
    for filename in files:
      tar.add(filename, recursive = True)
  finally: tar.close()
  output.set_metainfo(tar.recorded_files())

def extract_archive(input, files = [], verbose = False):
  tar = taropen(fileobj = input, mode = 'r|')
  try:
    member = tar.next()
    while member:
      if len(files) > 0:
        if member.name in files:
          tar.extract(member)
          if verbose:
            print member.name
            sys.stdout.flush()
      else:
        tar.extract(member)
        if verbose:
          print member.name
          sys.stdout.flush()
      member = tar.next()
  finally: tar.close()

def list_archive(input, files = [], verbose = False):
  for finfo in input.metainfo():
    if len(files) > 0 and finfo['name'] not in files: continue
    if verbose:
      print filemode(finfo['mode']),
      print "%- 15s" % ("%s/%s" % (finfo['uname'] or finfo['uid'], finfo['gname'] or finfo['gid'])),
      if finfo['type'] == CHRTYPE or finfo['type'] == BLKTYPE:
        print "%10s" % ("%d,%d" % (finfo['devmajor'], finfo['devminor'])),
      else: print "%10d" % finfo['size'],
      print "%d-%02d-%02d %02d:%02d:%02d" % time.localtime(finfo['mtime'])[:6],
    print finfo['name'],
    if verbose:
      if finfo['type'] == SYMTYPE: print "->", finfo['linkname'],
      if finfo['type'] == LNKTYPE: print "link to", finfo['linkname'],
    print

def search_archive(util, keywords, verbose = False):
  result, parents, errors = util.search(keywords)
  print "% 6s" % "WEIGHT",
  print "%- 43s" % "ARCHIVE",
  print " %- 17s" % "CREATED",
  print " %- 13s" % "SIZE",
  print " %- 13s" % "USAGE",
  print "DESCRIPTION"
  for res in result:
    if not verbose and parents.has_key(res['id']): continue
    print "% 6d" % res['weight'],
    print res['id'],
    print " " + time.strftime("%F %R", time.localtime(res['timestamp'])),
    print " %- 13d" % res['size'],
    print " %- 13d" % res['usage'],
    print " " + res['description']
  if verbose:
    for metadata, error in errors:
      print "WARNING:", str(error)

def list_keywords(util, verbose = False):
  keywords, errors = util.keywordlist()
  keywords = keywords.items()
  keywords.sort(key = lambda x: x[0])
  print "%- 30s %- 5s" % ("KEYWORD", "USAGE")
  for kwd, usage in keywords:
    print "%- 30s%- 5d" % (kwd, usage)
  if verbose:
    for metadata, error in errors:
      print "WARNING:", str(error)

def copy_archive(source_stor, dest_stor, archive, verbose = False):
  block_list = source_stor.blocks(archive)
  if verbose:
    sys.stdout.write("Copy %d blocks " % len(block_list))
    sys.stdout.flush()
  for number, checksum1, checksum2, length in block_list:
    if not dest_stor.block_exists(number, checksum1, checksum2):
      buffer = source_stor.read_buffer(number, checksum1, checksum2)
      dest_stor.write_buffer(buffer, number, checksum1, checksum2)
      if verbose: sys.stdout.write('.'); sys.stdout.flush()
    else:
      if verbose: sys.stdout.write('+'); sys.stdout.flush()
  if verbose:
    sys.stdout.write(' done.\n')

def verify_archive(util, archive, verbose = False):
  util.verify_blocks(archive)

def command_tar_arguments(args, config_files = []):
  optparser = ArgumentParser(
      fromfile_prefix_chars = '@',
      formatter_class = RawDescriptionHelpFormatter,
      description = """Store deduplicated, compressed and encrypted data failsafe distributed over
filesystems, sFTP accounts, IMAP mailboxes or WebDAV shares.
""",
      epilog = """

The nimbstor allows you to store archives in different places through
simple backends. It comes with tool nimbtar, which stores tar archives,
but with following additional features:

  * Store multiple archives with description and keywords
  * Search for description, keywords and filenames
  * List archive contents without reading it
  * Distribute the tar archive over several storage places
  * Use deduplication to store multiple copies of same data efficiently
  * Create redundancy to compensate fail of one or more storage places
  * Encrypt data with AES 256

Following backends are supported for now: filesystems, IMAP, WebDAV,
SFTP. To use it, you need to set the STORAGE parameter to following
valeus:

  Filesystem    <directory name>

  IMAP          imap:<host>:<port>:<user>:<password>:<mailbox>
                imaps:<host>:<port>:<user>:<password>:<mailbox>

  WebDAV        webdav:<host>:<port>:<user>:<password>:<directory>
                webdavs:<host>:<port>:<user>:<password>:<directory>

  SFTP          sftp:<host>:<port>:<user>:<password>:<directory>
                (You can use path to you private key as password)

On empty passwords nimbstor will ask one on the console.

If you specify multiple backends, then the archives will be distributed
over all backends and also the specified redundancy will be ccreated.
You can also specify PARALLEL parameter to parallalize the archivation
process.

Per default it will read the ~/.nimbtarrc and /etc/nimbtarrc config
files, if they are exists. This files can contain command line arguments
which you would preset. If environment variable NIMBTARCFG is set, then
the file specified there will be used instead.

ADDITIONAL REQUIREMENTS:

  Encryption              alo_aes
  LZMA compression        pyliblzma
  Parallelization         multiprocessing
                          (only for python 2.6 and less)
  WebDAV                  tinydav
  sFTP                    paramiko

You can install the requirements simply with pip or easy_install.

""")

  optparser.add_argument('-e', '--empty', action = 'store_true', help = 'Empty archive schould be written also.')
  optparser.add_argument('-v', '--verbose', action = 'store_true', help = 'Verbose mode.')

  optparser.add_argument('-d', '--storage', required = True, nargs = '+', help = 'Use given storages as archive.')
  optparser.add_argument('-D', '--desc', help = 'Description of the archive.')
  optparser.add_argument('-K', '--keywords', nargs = '+', metavar = 'KWD', help = 'List of keywords for archive.')

  optparser.add_argument('-a', '--archive', metavar = 'ID', help = 'Give the archive ID to use (as parent for creation).')
  optparser.add_argument('-r', '--redundancy', type = int, default = 1, help = 'Store each block this amount of times.')
  optparser.add_argument('-C', '--change-directory', metavar = 'DIR', help = 'Operate in given directory.')
  optparser.add_argument('-m', '--parallel', type = int, metavar = 'NUM', default = None, help = 'Count of workers to start (default: cpu count).')
  optparser.add_argument('-p', '--password', default = True, nargs = '?', metavar = 'PWD', help = 'Use this encryption password or ask for it.')

  optgroup2 = optparser.add_mutually_exclusive_group(required = False)
  optgroup2.add_argument('-z', '--gzip', type = int, default = None, metavar = 'LEVEL', help = 'Use GZip for compression.')
  optgroup2.add_argument('-j', '--bzip2', type = int, default = None, metavar = 'LEVEL', help = 'Use BZip2 for compression.')
  optgroup2.add_argument('-Z', '--lzma', type = int, default = None, metavar = 'LEVEL', help = 'Use LZMA for compression (default, level 9).')
  optgroup2.add_argument('-N', '--nocompress', action = 'store_true', help = 'Disable compression.')

  optgroup3 = optparser.add_mutually_exclusive_group(required = True)
  optgroup3.add_argument('-c', '--create', nargs = '+', metavar = 'FILE', help = 'Create an archive with given files.')
  optgroup3.add_argument('-x', '--extract', nargs = '*', metavar = 'FILE', help = 'Extract an archive fully or only given files.')
  optgroup3.add_argument('-t', '--list', nargs = '*', metavar = 'FILE', help = 'List an archive content fully or only given files.')
  optgroup3.add_argument('-s', '--search', nargs = '*', metavar = 'FILE', help = 'Search for an archive or list all archives.')
  optgroup3.add_argument('-k', '--list-keywords', action = 'store_true', help = 'List used keywords with usage count.')
  optgroup3.add_argument('--copy', nargs = '+', metavar = 'STORAGE', help = 'Copy an archive to other backend')
  optgroup3.add_argument('-F', '--verify', action = 'store_true', help = 'Download and verify given archive.')

  presets = []
  for cfgfile in config_files:
    if os.path.exists(cfgfile):
      for conf in file(cfgfile):
        presets.extend(optparser.convert_arg_line_to_args(conf.strip()))
  options = optparser.parse_args(presets + args[1:])

  if (options.extract or options.list) \
      and not options.archive:
    sys.stderr.write("Please give an archive (-a) ID to operate.\n")
    sys.exit(1)

  if options.change_directory:
    os.chdir(options.change_directory)

  if options.redundancy < 1 or options.redundancy > len(options.storage):
    sys.stderr.write("Redundancy need to be between 1 and %d.\n" % len(options.storage))
    sys.exit(1)

  if options.lzma != None: compression = ('lzma', options.lzma)
  elif options.gzip != None: compression = ('gzip', options.gzip)
  elif options.bzip2 != None: compression = ('bzip2', options.bzip2)
  elif options.nocompress: compression = (None, 0)
  else: compression = ('lzma', 9)

  if options.password == None:
    password = ''; verify = None
    while password != verify:
      password     = getpass.unix_getpass(prompt = 'Encryption password: ')
      if options.create == None: verify = password
      else: verify = getpass.unix_getpass(prompt = '             Verify: ')
  elif options.password == True: password = None
  else: password = options.password

  if compression[0] == None and password: parallel = None
  elif options.parallel == None: parallel = 1
  elif options.parallel < 1:
    sys.stderr.write("Parallization should be larger zero.\n")
    sys.exit(1)
  else: parallel = options.parallel

  return {
      'parser': optparser,
      'options': options,
      'compression': compression,
      'password': password,
      'parallel': parallel,
      'verbose': options.verbose,
      'redundancy': options.redundancy,
      'archive': options.archive,
      'storage': options.storage,
  }

def create_sub_backend(options, nimbstor, dir):
  if dir.startswith('imap:') or dir.startswith('imaps:'):
    proto, host, port, user, password, mailbox = dir.split(':', 6)
    port = int(port)
    from nimbstor.imap_backend import imap_backend
    return (imap_backend(nimbstor, user, password, host, mailbox, port,
                         proto == 'imaps', options['verbose']),
            "%s:%s:%d:%s::%s" % (proto, host, port, user, mailbox))
  elif dir.startswith('webdav:') or dir.startswith('webdavs:') :
    proto, host, port, user, password, basedir = dir.split(':', 6)
    port = int(port)
    if proto == 'webdav': urltype = 'http'
    else: urltype = 'https'
    from nimbstor.webdav_backend import webdav_backend
    return (webdav_backend(nimbstor, urltype, host, port, basedir, user, password, options['verbose']),
            "%s:%s:%s:%s::%s" % (proto, host, port, user, basedir))
  elif dir.startswith('sftp:'):
    proto, host, port, user, password, basedir = dir.split(':', 6)
    from nimbstor.sftp_backend import sftp_backend
    return (sftp_backend(nimbstor, host, port, basedir, user, password, options['verbose']),
            "%s:%s:%s:%s::%s" % (proto, host, port, user, basedir))
  from nimbstor.directory_backend import directory_backend
  return (directory_backend(nimbstor, dir, options['verbose']), dir)

def create_backend(options, nimbstor, storagelist):
  if len(storagelist) == 1:
    comb, name = create_sub_backend(options, nimbstor, storagelist[0])
    comb.activate()
    if options['verbose']:
        print "Use following backends:", name
    return comb
  namelist = []
  from nimbstor.combine_backend import combine_backend
  comb = combine_backend(nimbstor, options['redundancy'], options['verbose'])
  for dir in storagelist:
    x, n = create_sub_backend(options, comb, dir)
    x.activate()
    namelist.append(n)
  comb.activate()
  if options['verbose']:
    print "Use following backends:", ', '.join(namelist)
  return comb

def command_create_archive(options, start_time):
  """Create archive (tar -c)

  >>> os.system('rm -rf .nimbstor_parent out*/* || true')
  0
  >>> opt = command_tar_arguments(['test', '-v', '-p', 'qwer', '-m', '10', '-d', 'out1', 'out2', 'out3', '-D', 'test1', '-K', 'kwd1', '-c', 'f1', 'f2', 'f3'])
  >>> command_create_archive(opt, time.time())
  Use following backends: out1, out2, out3
  f1
  f2
  f3
  Archive ...

  """
  if not options['options'].desc:
    raise SystemError("Description must be given to create archive.")
  keywords = [x.lower().strip() for x in options['options'].keywords or []]
  desc = options['options'].desc.strip()
  output = stor_output(options['password'], options['compression'], options['archive'], desc, keywords, options['parallel'], not options['options'].empty, options['verbose'])
  dont_commit = True
  try:
    create_backend(options, output, options['storage'])
    if os.path.exists('.nimbstor_parent'):
      parent = open('.nimbstor_parent').read().strip()
      if len(parent) == 64:
        if options['verbose']:
          print "Use parent", parent
        output.set_parent(parent)
    create_archive(output, options['options'].create, options['options'].verbose)
    dont_commit = False
  finally: output.close(dont_commit)
  if output.archive() != None:
    open('.nimbstor_parent', 'w').write(output.archive())
  if options['verbose'] and output.archive() == None:
    print "No archive created."
  elif options['verbose']:
    print "Archive", output.archive(), "with size", output.size(), "created in", time.time() - start_time, "seconds."

def command_extract_archive(options, start_time):
  """Extract archive (tar -x)

  >>> archives = list(set([x[17:] for x in os.listdir('out1') if x.startswith('0000000.00000000.')]))
  >>> opt = command_tar_arguments(['test', '-m', '10', '-d', 'out1', 'out2', 'out3', '-p', 'qwer', '-v', '-a', archives[0], '-x'])
  >>> command_extract_archive(opt, time.time())
  Use following backends: out1, out2, out3
  f1
  f2
  f3
  Archive ...

  """
  input = stor_input(options['password'], options['archive'], options['parallel'], options['verbose'])
  try:
    create_backend(options, input, options['storage'])
    extract_archive(input, options['options'].extract, options['verbose'])
  finally: input.close()
  open('.nimbstor_parent', 'w').write(input.archive())
  if options['verbose']:
    print "Archive", options['archive'], "extracted in", time.time() - start_time, "seconds."

def command_list_archive(options, start_time):
  """List archive (tar -t)

  >>> archives = list(set([x[17:] for x in os.listdir('out1') if x.startswith('0000000.00000000.')]))
  >>> opt = command_tar_arguments(['test', '-d', 'out1', 'out2', 'out3', '-p', 'qwer', '-a', archives[0], '-t'])
  >>> command_list_archive(opt, time.time())
  f1
  f2
  f3

  """
  input = stor_input(options['password'], options['archive'], options['parallel'], options['verbose'])
  try:
    create_backend(options, input, options['storage'])
    if options['verbose']:
      print "ARCHIVE INFO:"
      print "  Description:    ", input.description()
      print "  Keywords:       ", input.keywords()
      print "  Parent:         ", input.parent()
      print "  Size:           ", input.size()
      print "  Usage:          ", input.usage()
      print "  Compression:    ", input.compression()
      print "-"
    list_archive(input, options['options'].list, options['verbose'])
    if options['verbose']:
      print "-"
      print "USED BLOCK NAME                                              SIZE"
      count = 0
      for num, checksum1, checksum2, blsize in input.blocks():
        t = '(unknown)'
        if blsize > 0: t = blsize
        print "%07d.%s.%s" % (num, checksum1, checksum2), t
        count = count + 1
      print "-"
      print "Archive have", count, "blocks."
  finally: input.close()

def command_copy_archive(options, start_time):
  """Copy archive

  >>> archives = list(set([x[17:] for x in os.listdir('out1') if x.startswith('0000000.00000000.')]))
  >>> opt = command_tar_arguments(['test', '-d', 'out3', '-p', 'qwer', '-a', archives[0], '-v', '--copy', 'out1', 'out2'])
  >>> command_copy_archive(opt, time.time())
  Use following backends: out3
  Use following backends: out1, out2
  Copy 2 blocks ...

  """
  input = stor_util(options['password'])
  try:
    create_backend(options, input, options['storage'])
    dest = stor_util(options['password'])
    dont_commit = True
    try:
      create_backend(options, dest, options['options'].copy)
      copy_archive(input, dest, options['archive'], options['verbose'])
      dont_commit = False
    finally: dest.close(dont_commit)
  finally: input.close()

def command_search_archive(options, start_time):
  """Search for archive

  >>> opt = command_tar_arguments(['test', '-d', 'out1', 'out2', 'out3', '-p', 'qwer', '-v', '-s'])
  >>> command_search_archive(opt, time.time())
  Use following backends: out1, out2, out3
  WEIGHT ARCHIVE                                      CREATED            SIZE           USAGE         DESCRIPTION
       0 ...

  """
  util = stor_util(options['password'])
  try:
    create_backend(options, util, options['storage'])
    search_archive(util, options['options'].search, options['verbose'])
  finally: util.close()

def command_list_keywords(options, start_time):
  """List used keywords.

  >>> opt = command_tar_arguments(['test', '-d', 'out1', 'out2', 'out3', '-p', 'qwer', '-v', '-k'])
  >>> command_list_keywords(opt, time.time())
  Use following backends: out1, out2, out3
  KEYWORD                        USAGE
  kwd1                           1   


  """
  util = stor_util(options['password'])
  create_backend(options, util, options['storage'])
  list_keywords(util, options['verbose'])

def command_verify_archive(options, start_time):
  util = stor_util(options['password'], options['verbose'])
  create_backend(options, util, options['storage'])
  verify_archive(util, options['archive'], options['verbose'])

__all__ = (
    'create_archive',
    'extract_archive',
    'list_archive',
    'search_archive',
    'list_keywords',
    'copy_archive',
    'command_tar_arguments',
    'command_create_archive',
    'command_extract_archive',
    'command_list_archive',
    'command_copy_archive',
    'command_search_archive',
    'command_list_keywords',
    'command_verify_archive',
)

# vim:sts=2:sw=2:



cdef unsigned long _character_table[256]
def _initialize():
    cdef unsigned int i = 0
    for n in [
0x272ab691, 0xcde74fe9, 0xf530e162, 0x96f9db97, 0x6150d109, 0x2e6ee7e3,
0xb191c9bb, 0xf898f264, 0x84daacc3, 0x16fcc550, 0x73ca0ba5, 0x4f30b942,
0xf459c982, 0xa2c7cf3e, 0x69d0eebf, 0xf945e6f0, 0x8d8c92e8, 0xf92a953b,
0x225300d0, 0xb1f45ef3, 0xa97278fc, 0x4ed65be1, 0xb91f3643, 0x3cf6860f,
0xa49a98ad, 0x10ab3945, 0x8ae79aa4, 0xd792a6d1, 0x266ef44b, 0x41c74186,
0x11379127, 0x3f206606, 0xb47e3c07, 0xf69ec345, 0x01c1a523, 0xc2184745,
0xfd06267e, 0x7f2f9425, 0x4e66ea90, 0xbcdf4162, 0xaa405927, 0x42a16d54,
0x2d99ee8b, 0x17c062f8, 0x7e33e0e4, 0x86e557fa, 0xe4ff37da, 0xd7c686e9,
0xed0d3a7b, 0x5f2629ca, 0xaa6d5760, 0x986818d5, 0x6a3eeb12, 0xc41c117b,
0x91ae88a9, 0xd036e9a5, 0x22eca8b6, 0x69953343, 0x27a5e263, 0x17061403,
0xee6312d7, 0x3cc55561, 0x0dba16de, 0x4037f0df, 0xd39dfac9, 0x972e9c95,
0xa2a7c492, 0x00261bec, 0xa11f3c43, 0x1a3558a6, 0x881d5699, 0xffd58820,
0xba9825e3, 0xd7b0e0da, 0x9ec27531, 0xb668a438, 0x6e2771a9, 0xe40b8083,
0xe48d063b, 0x1c018f32, 0xd92dcd79, 0xaec3736d, 0x7b85bdbf, 0x4e065cd0,
0xa2faf39a, 0x865ef270, 0x015e3e66, 0x88f96a1c, 0x99d55337, 0x7d254fe5,
0x4a7ac9fe, 0x7057c1df, 0x8d1e48e4, 0x7c2f2428, 0x855181fa, 0x792f9127,
0x9594e712, 0xb29f8a1a, 0xc54137f0, 0x7060abca, 0x5af94e9b, 0xb7186ba2,
0xf641ce19, 0x81b920e0, 0xcc92973a, 0xa1bbde27, 0x451027d4, 0xd95fdf40,
0x30594811, 0x1d4c8f17, 0x19b7aa23, 0xf53f6a0b, 0x62c86021, 0xee3f4064,
0xdc9b4c2c, 0x00629d23, 0x9f96eed2, 0xb6cf5501, 0xe06037b8, 0x1e7581b3,
0x31dc3c74, 0xcf686d6b, 0xf55bf9b9, 0x61dc32a2, 0xf86082b5, 0xe2f7141a,
0x5d6842bd, 0x4311d121, 0xc11c3527, 0x9ef27a12, 0x113e377b, 0xd4418923,
0x1d27d9c7, 0x2f314359, 0x8abb514e, 0x984ba5dd, 0x749e716c, 0x6ddb6377,
0x91b1ceb5, 0xd8a8b64e, 0xd034705a, 0xa80b09f9, 0xdb26b3cf, 0x91dcc1b3,
0x048546b2, 0x97ee9eee, 0xa51f5ae7, 0x926c8c10, 0xa95978ff, 0xbd7dcf6f,
0xdb6094e2, 0x3486eaa9, 0x8d6748ab, 0x8f13aa15, 0xbbc159d0, 0x2ed4288e,
0x254fec86, 0x137fcb7a, 0x62f30964, 0x551d7b09, 0xcadb5947, 0x57609ac0,
0x24e13b75, 0xc5be8632, 0x7fb49a03, 0xd7d8960d, 0x160b207b, 0xed12f15e,
0x4d28b8b9, 0x6cf0af0c, 0xb5787215, 0xfc520ba7, 0x48aa36c6, 0x7639b765,
0x392bf384, 0xed8a806c, 0xbb1b156f, 0x0c7edd4f, 0x9935822f, 0xf2b01bd4,
0xe71f6e5d, 0xc39b37fa, 0x17c5b621, 0x39ed9f40, 0x70fa9c1b, 0x17c46422,
0xaee51288, 0x45aca5e2, 0x78b8f29c, 0xf9dc3084, 0x7eb0c634, 0xb3059924,
0x0da807b0, 0x0e9b5a62, 0x113efc1d, 0x2d89b53b, 0xc05f8b92, 0x1ca759ff,
0x0b8cbbde, 0x7b991d9d, 0xbedcb15e, 0xd5e53f23, 0xad175a32, 0x02b70186,
0x14722792, 0xa09b9da0, 0x5203797d, 0xdf8dd077, 0x5d8ea095, 0x453589c3,
0x05f39a47, 0xb1957eaf, 0x7d9beb2d, 0x6db8a4e2, 0x17f03bb7, 0x868dd9b8,
0xab789b07, 0xb44039d6, 0x8a038c1c, 0xe63eecdd, 0x774e4132, 0x960f697c,
0x296aa969, 0xdb3175ac, 0x42c634a4, 0x010c6738, 0x167e71b4, 0xf951687b,
0x02deeab4, 0x59b6c56c, 0x85b5e3b6, 0xcfac84dc, 0x08e3df9f, 0xc7ead201,
0x8d640f4d, 0xe6befa2f, 0xfe4a0f03, 0x510402e6, 0xe61395fe, 0x69001c10,
0x3572fc65, 0xc6307289, 0x9da1bb6d, 0xd7a126e4, 0xef984bb0, 0xd6076c64,
0xdf7b05e2, 0xb9b01451, 0x3e61e9ad, 0xa0b5096c, 0x1e9b6ed0, 0x08107fc1,
0xb5e8c48c, 0x064cdca1, 0x20342345, 0x34fc9ae7 ]:
        _character_table[i] = n
        i = i + 1
_initialize()

def rolling_checksum(buf):
    cdef unsigned long cksum = 0
    cdef unsigned long i = 0
    cdef unsigned long buflen = len(buf)
    cdef unsigned char* chrbuf = buf

    while i < buflen:
        cksum = cksum ^ _character_table[chrbuf[i]]
        i = i + 1

    return cksum

def rolling_search(dic, checkfunc, buf, blocksize):
    cdef unsigned long cksum = 0
    cdef unsigned long fcks = 0
    cdef unsigned long i = 0 
    cdef unsigned long pos = 0 
    cdef unsigned long end = 0 
    cdef unsigned long buflen = len(buf)
    cdef unsigned long bs = blocksize
    cdef unsigned char* chrbuf = buf

    while i < bs:
        cksum = cksum ^ _character_table[chrbuf[i]]
        i = i + 1

    fcks = cksum ; end = buflen - bs - 1
    while pos < end:
        if dic.has_key(cksum):
            if checkfunc(chrbuf[pos:pos+bs], pos, cksum):
                return (pos, cksum)
        cksum = cksum ^ _character_table[chrbuf[pos]]
        cksum = cksum ^ _character_table[chrbuf[pos+bs]]
        pos = pos + 1
    return (None, fcks)

class stor_buffer:
    """Manage buffer for nimbstor.
   
    Not a performant implementation, should be optimized one day.
   
    """
   
    def __init__(self, data = "", splitsize = None):
        """Initialize the buffer.
       
        Args:
          data (str): Preset buffer content.
       
        >>> stor_buffer("asdf").value()
        'asdf'
        """
        self._data = [data]
        self._len = len(data)
        self._val = None
        self._splitsize = splitsize
   
    def __len__(self):
        """Return buffer content length.
       
        >>> len(stor_buffer("asdfqw"))
        6
        """
        return self._len
   
    def push(self, data, left = False):
        """Push data to the buffer.
       
        Args:
          data (str): Data to be pushed.
          left (bool): Push to begin of the buffer (default: False)
       
        >>> buf = stor_buffer('asdf')
        >>> buf.push('qwer')
        >>> buf.push('zxcv', True)
        >>> buf.value()
        'zxcvasdfqwer'
        """
        self._len += len(data)
        if self._splitsize != None:
            listdat = []
            curpos = 0
            while curpos < self._len:
                listdat.append(data[curpos:curpos + self._splitsize])
                curpos += self._splitsize
            if left: self._data[0:0] = listdat
            else: self._data.extend(listdat)
        else:
            if left: self._data.insert(0, data)
            else: self._data.append(data)
        self._val = None
   
    def _rmsize(self, size):
      """Internal function to remove some data from buffer.
   
      Args:
        size (int): Amount of data to remove.
   
      This function expects, that self._val has correct value.
   
      >>> buf = stor_buffer('asdfqwer')
      >>> buf.value()
      'asdfqwer'
      >>> buf._rmsize(3)
      >>> buf.value()
      'fqwer'
      """
      new = self._val[size:]
      self._val = None
      self._data = [new]
      self._len = len(new)
   
    def pop(self, size = None):
      """Pop some data from buffer.
   
      Args:
        size (int): Amount of data to pop.
   
      >>> buf = stor_buffer('asdfqwer')
      >>> buf.pop(3)
      'asd'
      >>> buf.value()
      'fqwer'
      """
      if size == None:
          buf = "".join(self._data)
          self._data = []
          self._len = 0
          self._val = None
          return buf
      if len(self._data) == 0:
          return ''
      self._val = None
      buf = []
      restsize = size
      while restsize > 0 and len(self._data) > 0:
          curbuf = self._data[0]
          curlen = len(curbuf)
          if curlen > restsize:
              buf.append(curbuf[:restsize])
              self._data[0] = curbuf[restsize:]
              self._len -= restsize
              restsize = 0
          else:
              buf.append(curbuf)
              del self._data[0]
              self._len -= curlen
              restsize -= curlen
      return ''.join(buf)
   
    def truncate(self, size = None):
      """Remove some amount of data from buffer.
   
      Args:
        size (int): Amount of data to remove (default: all).
   
      >>> buf = stor_buffer('asdfqwer')
      >>> buf.truncate(3)
      >>> buf.value()
      'fqwer'
      """
      if size == None:
          self._data = []
          self._val = None
          self._len = 0
          return
      if self._val == None: self._val = "".join(self._data)
      self._rmsize(size)
   
    def value(self):
      """Return the buffer value.
   
      >>> stor_buffer('asdf').value()
      'asdf'
      """
      if self._val == None: self._val = "".join(self._data)
      return self._val
   
# vim:sts=4:sw=4:



#!/usr/pkg/bin/python2.6 -O

import sys, random

class combine_backend:
    def __init__(self, nimbstor, redundancy = 1, verbose = False):
        self._nimbstor = nimbstor
        self._backends = []
        self._block_directory = {}
        self._redundancy = redundancy
        self._verbose = verbose

    def open(self, backend):
        self._backends.append(backend)

    def activate(self):
        self._nimbstor.open(self)

    def close(self, dont_commit = False):
        for backend in self._backends:
            backend.close(dont_commit)

    def append_block_info(self, number, checksum1, checksum2, backend):
        self._block_directory.setdefault(
                (number, checksum1, checksum2), []
        ).append(backend)
        self._nimbstor.append_block_info(number, checksum1, checksum2, self)
  
    def write_buffer(self, buf, number, checksum1, checksum2):
        written_count = 0
        write_backends = self._backends[:]
        random.shuffle(write_backends)
        if number == 0: redundancy = len(write_backends)
        else: redundancy = self._redundancy
        while written_count < redundancy:
            if len(write_backends) == 0 and written_count == 0:
                raise SystemError("Could not write to destination backends")
            elif len(write_backends) == 0 and written_count > 0:
                sys.stderr.write("Could only write %d blocks for %s\n" % \
                        (written_count, repr((number, checksum1, checksum2))))
                break
            backend = write_backends.pop()
            backend.write_buffer(buf, number, checksum1, checksum2)
            written_count += 1
  
    def read_buffer(self, number, checksum1, checksum2):
        read_backends = self._block_directory[(number, checksum1, checksum2)]
        random.shuffle(read_backends)
        for backend in read_backends:
            try: data = backend.read_buffer(number, checksum1, checksum2)
            except: continue
            return data
        raise SystemError("Failed to read block: " \
                + repr((number, checksum1, checksum2)))
  
__all__ = ('combine_backend',)



import imaplib, re, time, base64, getpass

re_subject = re.compile(r'^Subject:.* ([0-9]+)\.([0-9A-Fa-f]+)\.([0-9A-Za-z_-]+)')

class imap_backend:
    def __init__(self, nimbstor, user, password, host, mailbox = 'INBOX', port = 143, ssl = False, cram = False, verbose = False):
        self._nimbstor = nimbstor
        self._mailbox = mailbox
        if password == '':
            password = getpass.unix_getpass(prompt = '      IMAP password: ')
        self._connect = (user, password, host, port, ssl, cram)
        self._login()
        self._imap = None
        self._messages = {}
        self._verbose = verbose
        _, msglist = self._imap.search(None, 'ALL')
        for msgnum in msglist[0].split():
            _, header = self._imap.fetch(msgnum, '(RFC822.HEADER)')
            for headline in header[0][1].split('\r\n'):
                ma_subject = re_subject.match(headline)
                if ma_subject:
                    number = int(ma_subject.group(1))
                    checksum1 = ma_subject.group(2)
                    checksum2 = ma_subject.group(3)
                    self._messages[(number, checksum1, checksum2)] = msgnum
                    self._nimbstor.append_block_info(number, checksum1, checksum2, self)
                    break
  
    def _login(self):
        user, password, host, port, ssl, cram = self._connect
        if ssl: self._imap = imaplib.IMAP4_SSL(host, port)
        else: self._imap = imaplib.IMAP4(host, port)
        if cram: self._imap.login_cram_md5(user, password)
        else: self._imap.login(user, password)
        self._imap.select(self._mailbox)
  
    def activate(self):
        self._nimbstor.open(self)
  
    def close(self, dont_commit = False):
        pass
  
    def write_buffer(self, buf, number, checksum1, checksum2):
        message = '\r\n'.join([
          "Subject: %07d.%s.%s" % (number, checksum1, checksum2),
          '',
          base64.encodestring(buf),
        ])
        self._imap.append(self._mailbox, '', time.time(), message)
        self._nimbstor.append_block_info(number, checksum1, checksum2, self)
  
    def read_buffer(self, number, checksum1, checksum2, decode_func):
        msgnum = self._messages[(number, checksum1, checksum2)]
        _, message = self._imap.fetch(msgnum, '(RFC822.TEXT)')
        return decode_func(base64.decodestring(message[0][1]))

__all__ = ('imap_backend',)



#!/usr/pkg/bin/python2.6 -O

import os, tinydav, re, getpass

class webdav_backend:
    def __init__(self, nimbstor, proto, host, port, basedir, user, password, verbose = False):
        if basedir.endswith('/'): self._basedir = basedir
        else: self._basedir = basedir + '/'
        self._nimbstor = nimbstor
        self._verbose = verbose
        self._connection = tinydav.WebDAVClient(host, int(port), proto)
        if password == '':
            password = getpass.unix_getpass(prompt = '    WebDAV password: ')
        self._connection.setbasicauth(user, password)
        re_fname = re.compile(r'.*/([0-9]{7})\.([0-9A-Fa-f]+)\.([0-9A-Za-z_-]+)$')
        for fname in self._connection.propfind(self._basedir, depth = 1):
            ma_fname = re_fname.match(fname.href)
            if not ma_fname: continue
            number = int(ma_fname.group(1))
            checksum1 = ma_fname.group(2)
            checksum2 = ma_fname.group(3).upper()
            #if self._verbose: print "WebDAV FOUND:", fname.href
            self._nimbstor.append_block_info(number, checksum1, checksum2, self)
   
    def activate(self):
        self._nimbstor.open(self)
   
    def write_buffer(self, buffer, number, checksum1, checksum2):
        fname = os.path.join(self._basedir, "%07d.%s.%s" % (number, checksum1, checksum2))
        #if self._verbose: print "WebDAV PUT:", fname
        self._connection.put(fname, buffer)
        self._nimbstor.append_block_info(number, checksum1, checksum2, self)
   
    def read_buffer(self, number, checksum1, checksum2):
        fname = os.path.join(self._basedir, "%07d.%s.%s" % (number, checksum1, checksum2))
        #if self._verbose: print "WebDAV GET:", fname
        ret = self._connection.get(fname)
        return ret.content
   
    def close(self, dont_commit = False):
        pass

__all__ = ('directory_backend',)
# vim:sts=4:sw=4:



#!/usr/pkg/bin/python2.6 -O

import os

def read_directory_content(directory):
    content = []
    for fname in os.listdir(directory):
        number, checksum1, checksum2 = fname.split('.')
        number = int(number)
        content.append((number, checksum1, checksum2))
    return content

class directory_backend:
    def __init__(self, nimbstor, directory, verbose):
        self._directory = directory
        self._nimbstor = nimbstor
        self._verbose = verbose
        for number, checksum1, checksum2 in read_directory_content(directory):
            self._nimbstor.append_block_info(number, checksum1, checksum2, self)
  
    def activate(self):
        self._nimbstor.open(self)
  
    def write_buffer(self, buf, number, checksum1, checksum2):
        fname = os.path.join(self._directory, "%07d.%s.%s" % (number, checksum1, checksum2))
        if os.path.exists(fname):
            raise SystemError("DB: INTERNAL ERROR - block %s already exists" % fname)
        fd = open(fname, 'w')
        fd.write(buf)
        fd.close()
        # if self._verbose: print "DB: written", fname
        self._nimbstor.append_block_info(number, checksum1, checksum2, self)
  
    def read_buffer(self, number, checksum1, checksum2):
        fname = os.path.join(self._directory, "%07d.%s.%s" % (number, checksum1, checksum2))
        if not os.path.exists(fname):
            raise SystemError("DB: INTERNAL ERROR - block %s not exists" % fname)
        fd = open(fname, 'r')
        buf = fd.read()
        fd.close()
        # if self._verbose: print "DB: readen", fname
        return buf
  
    def close(self, dont_commit = False):
        pass
  
__all__ = ('directory_backend',)
# vim:sts=4:sw=4:



#!/usr/pkg/bin/python2.6 -O

import os, re, getpass
from paramiko import SFTPClient, Transport, RSAKey, DSSKey, SSHException

class sftp_backend:
    def __init__(self, nimbstor, host, port, basedir, user, password, verbose = False):
        self._nimbstor = nimbstor
        self._basedir = basedir
        self._verbose = verbose
        if password == '':
            password = getpass.unix_getpass(prompt = '      sFTP password: ')
        elif os.path.exists(os.path.expanduser(password)):
            try: password = RSAKey.from_private_key_file(os.path.expanduser(password))
            except SSHException:
                try: password = DSSKey.from_private_key_file(os.path.expanduser(password))
                except SSHException: pass
        self._transport = Transport((str(host), int(port)))
        if type(password) == str:
            self._transport.connect(username = user, password = password)
        else: self._transport.connect(username = user, pkey = password)
        self._connection = SFTPClient.from_transport(self._transport)
        self._connection.chdir(self._basedir)
        re_fname = re.compile(r'^([0-9]{7})\.([0-9A-Fa-f]+)\.([0-9A-Za-z_-]+)$')
        for fname in self._connection.listdir():
            ma_fname = re_fname.match(fname)
            if not ma_fname: continue
            number = int(ma_fname.group(1))
            checksum1 = ma_fname.group(2)
            checksum2 = ma_fname.group(3).upper()
            #if self._verbose: print "SFTP FOUND:", fname
            self._nimbstor.append_block_info(number, checksum1, checksum2, self)
  
    def activate(self):
        self._nimbstor.open(self)
  
    def write_buffer(self, buf, number, checksum1, checksum2):
        fname = "%07d.%s.%s" % (number, checksum1, checksum2)
        #if self._verbose: print "SFTP PUT:", fname
        fd = self._connection.open(fname, 'w')
        try: fd.write(buf)
        finally: fd.close()
        self._nimbstor.append_block_info(number, checksum1, checksum2, self)
  
    def read_buffer(self, number, checksum1, checksum2):
        fname = "%07d.%s.%s" % (number, checksum1, checksum2)
        #if self._verbose: print "SFTP GET:", fname
        fd = self._connection.open(fname, 'r')
        try: buf = fd.read()
        finally: fd.close()
        return buf
  
    def close(self, dont_commit = False):
        self._connection.close()

# vim:sts=4:sw=4:



#!/usr/bin/env python2.6

import sys, os
from setuptools import setup
from distutils.extension import Extension
from distutils.sysconfig import get_python_inc

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

dependencies = []
if sys.version_info.major < 3 and sys.version_info.minor < 7:
    dependencies.append('argparse')

import sys
if 'setuptools.extension' in sys.modules:
    m = sys.modules['setuptools.extension']
    m.Extension.__dict__ = m._Extension.__dict__

setup(
    name = "nimbstor",
    version = "1.0.1",
    description = "Store incrementally, compressed and encrypted data failsafe in filesystems, sFTP, WebDAV, IMAP or other backends.",
    author = "Oleksandr Kozachuk",
    author_email = "ddeus.pypi@mailnull.com",
    url = "http://pypi.python.org/pypi/nimbstor",
    license = "WTFPL",
    packages = ['nimbstor',],
    long_description = read('README'),
    install_requires = dependencies,
    setup_requires = ['setuptools_cython'],
    scripts = ["nimbstor/nimbtar"],
    ext_modules = [Extension("nimbstor.optimized", ["nimbstor/optimized.pyx"])],
    classifiers = [
      'Development Status :: 4 - Beta',
      'Environment :: Console',
      'Intended Audience :: End Users/Desktop',
      'Intended Audience :: System Administrators',
      'Operating System :: Unix',
      'Operating System :: POSIX',
      'Programming Language :: Python :: 2.6',
      'Programming Language :: Python :: 2.7',
      'Topic :: System :: Archiving',
      'Topic :: System :: Archiving :: Backup',
      'Topic :: System :: Archiving :: Compression',
      'Topic :: System :: Archiving :: Mirroring',
      'Topic :: System :: Archiving :: Packaging',
      'Topic :: Utilities',
    ],
    test_suite = 'nimbstor.tests.test_all',
    zip_safe = False,
)
